from django.db import models
from django.contrib.auth.models import Permission
from django.utils.translation import ugettext_lazy as _


class Transition(models.Model):

    name = models.CharField(_(u"Name"), max_length=100)
    destination = models.ForeignKey(
        "State",
        verbose_name=_(u"Destination"),
        related_name="destination_state")
    permission = models.ForeignKey(
        Permission,
        verbose_name=_(u"Permission"), blank=True, null=True)
    state = models.ForeignKey("State")

    def __unicode__(self):

        return self.name

    class Meta:

        ordering = ("name", )
        app_label = "djinn_workflow"



from django.template import Library
from djinn_workflow.utils import get_state


register = Library()


@register.filter(name="state")
def state(obj):

    try:
        return get_state(obj).name
    except:
        return ""



from django.test.testcases import TestCase
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from djinn_workflow.utils import (
    assign_workflow, get_workflow, set_state, get_state, apply_transition)
from djinn_workflow.models import Workflow
from djinn_workflow.signals import state_change


class UtilsTest(TestCase):

    def setUp(self):

        self.wf = Workflow.objects.create(name="test")

        active = self.wf.state_set.create(name="active")
        inactive = self.wf.state_set.create(name="inactive")

        inactive.transition_set.create(name="make_active", destination=active)
        active.transition_set.create(name="make_inactive",
                                     destination=inactive)

        self.wf.initial_state = inactive

    def test_assign_workflow(self):

        ct = ContentType.objects.get_for_model(User)

        assign_workflow(self.wf, ct)

        self.assertEquals(self.wf, get_workflow(ct))
        self.assertEquals(self.wf, get_workflow(User))

        user = User.objects.create(username="bobdobalina")

        self.assertEquals(self.wf, get_workflow(user))

    def test_get_state(self):

        ct = ContentType.objects.get_for_model(User)

        assign_workflow(self.wf, ct)

        user = User.objects.create(username="bobdobalina")

        self.assertEquals(None, get_state(user))

        set_state(user, "active")

        self.assertEquals("active", get_state(user).name)

    def test_apply_transition(self):

        ct = ContentType.objects.get_for_model(User)

        assign_workflow(self.wf, ct)

        user = User.objects.create(username="bobdobalina")

        set_state(user, "active")

        self.assertTrue(user.is_active)

        def callback(sender, instance, state_from, state_to, **kwargs):

            if state_to.name == "inactive":
                instance.is_active = False
                instance.save()

        state_change.connect(callback)

        apply_transition(user, "make_inactive")

        self.assertFalse(user.is_active)

        self.assertEquals("inactive", get_state(user).name)



from django.contrib.contenttypes.models import ContentType
from djinn_workflow.models.modelworkflow import ModelWorkflow
from djinn_workflow.models.workflow import Workflow
from djinn_workflow.models.state import State
from djinn_workflow.models.objectstate import ObjectState
from djinn_workflow.signals import state_change


def assign_workflow(wf, ct):

    """ Assign workflow to content type """

    ModelWorkflow.objects.get_or_create(workflow=wf, content_type=ct)


def get_workflow(anything):

    """ arg should be a model, an instance or a contenttype instance """

    if not isinstance(anything, ContentType):
        anything = ContentType.objects.get_for_model(anything)

    try:
        return ModelWorkflow.objects.get(content_type=anything).workflow
    except ModelWorkflow.DoesNotExist:
        return Workflow.objects.get(is_default=True)


def get_default_workflow():

    return Workflow.objects.get(is_default=True)


def set_state(obj, state):

    """ Set obj state, without doing the actual transition. This is mainly
    useful for setting initial state, i.e. during creation """

    wf = get_workflow(obj)

    if type(state) in [str, unicode]:
        state = wf.state_set.get(name=state)

    if not get_state(obj):
        ObjectState.objects.create(
            state=state,
            object_id=obj.id,
            object_ct=ContentType.objects.get_for_model(obj)
        )
    else:
        ObjectState.objects.filter(
            object_id=obj.id,
            object_ct=ContentType.objects.get_for_model(obj)
        ).update(state=state)


def get_state(obj, assume_initial=True, fail_silently=True):

    """ If fail_siltently is True, the get_state call will return
    an empty state """

    state = None

    try:
        state = ObjectState.objects.get(
            object_id=obj.pk,
            object_ct=ContentType.objects.get_for_model(obj)).state
    except ObjectState.DoesNotExist:
        if assume_initial:

            try:
                wf = get_workflow(obj)

                obj_state = ObjectState.objects.create(
                    state=wf.initial_state,
                    object_id=obj.id,
                    object_ct=ContentType.objects.get_for_model(obj)
                )

                state = obj_state.state
            except:
                if fail_silently:
                    state = State(name="no_state")

    return state


def apply_transition(obj, trans):

    """ Apply transitions, but this is unchecked for permissions! """

    state = get_state(obj)

    if type(trans) in [str, unicode]:
        trans = state.transition_set.get(name=trans)

    set_state(obj, trans.destination)

    state_change.send(obj.__class__, instance=obj,
                      state_from=state,
                      state_to=trans.destination)



from workflow import Workflow
from state import State
from transition import Transition
from objectstate import ObjectState
from modelworkflow import ModelWorkflow
from permissionassignment import PermissionAssignment

import signal_processors



from django.db.models.signals import post_save
from django.dispatch import receiver
from djinn_workflow.models.workflow import Workflow


@receiver(post_save, sender=Workflow)
def workflow_post_save(sender, instance, **kwargs):

    if instance.is_default:
        Workflow.objects.all().exclude(pk=instance.pk).update(is_default=0)



from django.test.testcases import TestCase
from djinn_workflow.models.workflow import Workflow


class WorkflowTest(TestCase):

    def setUp(self):

        pass

    def test_set_default(self):

        wf0 = Workflow.objects.create(name="test0")
        wf1 = Workflow.objects.create(name="test1")
        wf2 = Workflow.objects.create(name="test2")

        wf0.is_default = True
        wf0.save()

        self.assertTrue(wf0.is_default)
        self.assertFalse(wf1.is_default)
        self.assertFalse(wf2.is_default)

        wf1.is_default = True
        wf1.save()

        self.assertFalse(Workflow.objects.get(pk=wf0.pk).is_default)
        self.assertFalse(Workflow.objects.get(pk=wf2.pk).is_default)

        wf2.is_default = True
        wf2.save()

        self.assertFalse(Workflow.objects.get(pk=wf0.pk).is_default)
        self.assertFalse(Workflow.objects.get(pk=wf1.pk).is_default)



from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from djinn_workflow.models.state import State


class ObjectState(models.Model):

    object_ct = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    content_object = generic.GenericForeignKey('object_ct', 'object_id')
    state = models.ForeignKey(State)

    def __unicode__(self):
        return "%s - %s" % (self.content_object, self.state)

    class Meta:
        app_label = "djinn_workflow"
        unique_together = ("object_ct", "object_id")



from django.db import models
from django.utils.translation import ugettext_lazy as _


class State(models.Model):

    name = models.CharField(_(u"Name"), max_length=100)
    workflow = models.ForeignKey("Workflow")

    def __unicode__(self):

        return "%s" % self.name

    def get_transitions(self, obj, user):

        """Returns all allowed transitions for passed object and user.
        """

        transitions = []

        for transition in self.transition_set.all():

            permission = transition.permission

            if permission is None:
                transitions.append(transition)
            elif user.has_perm(permission.codename, obj=obj):
                transitions.append(transition)

        return transitions

    class Meta:

        ordering = ("name", )
        app_label = "djinn_workflow"



from django.db import models
from django.utils.translation import ugettext_lazy as _


class Workflow(models.Model):

    name = models.CharField(_(u"Name"), max_length=100, unique=True)
    initial_state = models.ForeignKey("State", related_name="initial_state",
                                      null=True, blank=True)
    is_default = models.BooleanField(default=False)

    def __unicode__(self):

        return self.name

    class Meta:

        ordering = ("name", )
        app_label = "djinn_workflow"



from django.db import models
from django.contrib.auth.models import Permission
from djinn_auth.models.role import Role


class PermissionAssignment(models.Model):

    """ Permission/role assignment bound to a state
    """

    state = models.ForeignKey("State")
    role = models.ForeignKey(Role)
    permission = models.ForeignKey(Permission)

    class Meta:

        app_label = "djinn_workflow"

    def __unicode__(self):

        return u"%s for %s in state %s" % (self.permission, self.role,
                                           self.state)



from django.db import models
from django.contrib.contenttypes.models import ContentType
from djinn_workflow.models.workflow import Workflow


class ModelWorkflow(models.Model):

    content_type = models.ForeignKey(ContentType, unique=True)
    workflow = models.ForeignKey(Workflow)

    def __unicode__(self):
        return "%s - %s" % (self.content_type.name, self.workflow.name)

    class Meta:

        app_label = "djinn_contenttypes"
